package org.basex.query;

import static org.basex.core.Text.*;
import org.basex.core.Main;
import org.basex.io.IO;
import org.basex.query.iter.Iter;
import org.basex.util.InputInfo;
import org.basex.util.InputParser;
import org.basex.util.StringList;
import org.basex.util.Token;

/**
 * This class indicates exceptions during query parsing or evaluation.
 *
 * @author Workgroup DBIS, University of Konstanz 2005-10, ISC License
 * @author Christian Gruen
 */
public final class QueryException extends Exception {
  /** Error items. */
  public Iter iter = Iter.EMPTY;

  /** File reference. */
  private IO file;
  /** Possible completions. */
  private StringList complete;
  /** Error code. */
  private String code;
  /** Error line. */
  private int line;
  /** Error column. */
  private int col;

  /**
   * Constructor.
   * @param ii input info
   * @param s message
   * @param e message extension
   */
  public QueryException(final InputInfo ii, final Object s, final Object... e) {
    super(Main.info(s, chop(e)));
    if(ii == null) return;

    line = 1;
    col = 1;
    final int qp = Math.min(ii.pos - 1, ii.query.length());
    for(int i = 0; i < qp; i++) {
      final char ch = ii.query.charAt(i);
      if(ch == 0x0A) { line++; col = 1; } else if(ch != 0x0D) { col++; }
    }
  }

  /**
   * Constructor.
   * @param ii input info
   * @param s xquery error
   * @param e error arguments
   */
  public QueryException(final InputInfo ii, final Object[] s,
      final Object... e) {
    this(ii, s[2], e);
    code = s[1] == null ? s[0].toString() : String.format("%s%04d", s[0], s[1]);
  }

  /**
   * Chops the specified array entries to a maximum length.
   * @param t token array
   * @return argument
   */
  private static Object[] chop(final Object[] t) {
    for(int i = 0; i < t.length; i++) {
      if(t[i] instanceof byte[]) {
        t[i] = Token.string((byte[]) t[i]);
      } else if(t[i] instanceof Throwable) {
        final Throwable th = (Throwable) t[i];
        t[i] = th.getMessage() != null ? th.getMessage() : th.toString();
      } else if(!(t[i] instanceof String)) {
        t[i] = t[i].toString();
      }
      // [CG] XQuery/Exception: verify if/which strings are to be chopped
      //final String s = t[i].toString();
      //t[i] = s.length() > 1000 ? s.substring(0, 1000) + DOTS : s;
    }
    return t;
  }
  
  /**
   * Returns the error code.
   * @return position
   */
  public String code() {
    return code;
  }

  /**
   * Returns the error column.
   * @return error column
   */
  public int col() {
    return col;
  }

  /**
   * Returns the error line.
   * @return error line
   */
  public int line() {
    return line;
  }

  /**
   * Possible completions.
   * @return error line
   */
  public StringList complete() {
    return complete == null ? new StringList() : complete;
  }

  /**
   * Finds line and column for the specified query parser.
   * @param parser parser
   */
  void pos(final InputParser parser) {
    // check if information has already been added
    if(line != 0) return;

    file = parser.file;
    line = 1;
    col = 1;
    for(int i = 0; i < parser.qm && i < parser.ql; i++) {
      final char ch = parser.qu.charAt(i);
      if(ch == 0x0A) { line++; col = 1; } else if(ch != 0x0D) { col++; }
    }
  }

  /**
   * Sets code completions.
   * @param qp query parser
   * @param comp completions
   */
  public void complete(final InputParser qp, final StringList comp) {
    complete = comp;
    pos(qp);
  }

  /**
   * Returns the simple error message.
   * @return string
   */
  public String simple() {
    return super.getMessage();
  }

  /**
   * Returns an extended error message.
   * @return string
   */
  public String extended() {
    final StringBuilder sb = new StringBuilder();
    if(code != null) sb.append("[" + code + "] ");
    return sb + simple();
  }

  @Override
  public String getMessage() {
    final StringBuilder sb = new StringBuilder();
    if(line != 0) {
      sb.append(STOPPED + ' ');
      sb.append(Main.info(LINEINFO, line));
      if(col != 0) sb.append(", " + Main.info(COLINFO, col));
      if(file != null) sb.append(Main.info(' ' + FILEINFO, file));
      sb.append(": \n");
    }
    return sb.append(extended()).toString();
  }
}
