package org.basex.query.expr;

import static org.basex.query.QueryText.*;
import static org.basex.query.QueryTokens.*;
import java.io.IOException;
import org.basex.data.Serializer;
import org.basex.query.QueryContext;
import org.basex.query.QueryException;
import org.basex.query.item.QNm;
import org.basex.query.item.SeqType;
import org.basex.query.item.Value;
import org.basex.query.iter.Iter;
import org.basex.query.iter.ItemIter;
import org.basex.util.InputInfo;
import org.basex.util.Token;

/**
 * Function call.
 *
 * @author Workgroup DBIS, University of Konstanz 2005-10, ISC License
 * @author Christian Gruen
 */
public final class FuncCall extends Arr {
  /** Function id reference. */
  private final int id;
  /** Function reference. */
  private Func func;
  /** Function name. */
  private final QNm name;

  /**
   * Function constructor.
   * @param ii input info
   * @param n function name
   * @param d function id
   * @param a arguments
   */
  public FuncCall(final InputInfo ii, final QNm n, final int d,
      final Expr[] a) {
    super(ii, a);
    id = d;
    name = n;
  }

  @Override
  public Expr comp(final QueryContext ctx) throws QueryException {
    func = ctx.fun.get(id);
    super.comp(ctx);

    // inline if result and arguments are all values.
    // [CG] XQuery/Inlining:
    if(func.expr.value() && values()) {
      // evaluate arguments to catch cast exceptions
      for(int a = 0; a < expr.length; a++) func.args[a].bind(expr[a], ctx);
      ctx.compInfo(OPTINLINE, func.var.name.atom());
      return func.iter(ctx).finish();
    }
    return this;
  }

  @Override
  public Iter iter(final QueryContext ctx) throws QueryException {
    if(func == null) {
      final Expr e = comp(ctx);
      if(e != this) return e.iter(ctx);
    }
    
    final int al = expr.length;
    final Value[] args = new Value[al];
    // evaluate arguments
    for(int a = 0; a < al; a++) args[a] = ctx.iter(expr[a]).finish();
    // move variables to stack
    final int s = ctx.vars.size();
    for(int a = 0; a < al; a++) {
      ctx.vars.add(func.args[a].bind(args[a], ctx).copy());
    }
    // evaluate function and reset variable scope
    final ItemIter ir = ItemIter.get(ctx.iter(func));
    ctx.vars.reset(s);
    return ir;
  }

  @Override
  public SeqType returned(final QueryContext ctx) {
    return func.returned(ctx);
  }

  @Override
  public boolean uses(final Use u, final QueryContext ctx) {
    return u == Use.UPD ? (func == null ? ctx.fun.get(id) : func).updating :
      super.uses(u, ctx);
  }

  @Override
  public void plan(final Serializer ser) throws IOException {
    ser.openElement(this, NAM, Token.token(toString()));
    for(final Expr e : expr) e.plan(ser);
    ser.closeElement();
  }

  @Override
  public String desc() {
    return "Function";
  }

  @Override
  public String toString() {
    return Token.string(name.atom()) + PAR;
  }
}
